from typing import List, Union, Dict

from ..event_types import EventType, Event, HistoryEntry, LogEntry


EventOrId = Union[EventType, int]


class EventManager:
    """
    Implementation-agnostic controller for events

    A list of already existing events can be manipulated by an instance of
    this class out of the box, but the actual IO is done by subclasses.
    """
    events: Dict[int, Event]

    class ManagerException(Exception):
        pass

    def __init__(self, session=None):
        self.session = session
        self.events = {}
        self.removed_ids = set()

    def _get_event(self, event_or_id: EventOrId) -> Event:
        if isinstance(event_or_id, Event):
            return event_or_id
        if isinstance(event_or_id, int):
            return self.events[event_or_id]
        raise ValueError("Unknown type")

    def _get_event_id(self, event_or_id: EventOrId) -> int:
        if isinstance(event_or_id, int):
            return event_or_id
        if isinstance(event_or_id, Event):
            return event_or_id.id
        raise ValueError("Unknown type")

    def _set_event(self, event: Event):
        self.events[event.id] = event

    def remove_event(self, event_or_id: EventOrId):
        event_id = self._get_event_id(event_or_id)
        self.events.pop(event_id, None)
        self.removed_ids.add(event_id)

    def _verify_session(self, quiet=False):
        if not self.session:
            if quiet:
                return False
            raise ValueError  # raise correct error
        return True

    def set_history_for_event(self, event_or_id: EventOrId, history_list: List[HistoryEntry]) -> Event:
        event = self._get_event(event_or_id)
        event.history = history_list
        self._set_event(event)
        return event

    def set_log_for_event(self, event_or_id: EventOrId, log_list: List[LogEntry]) -> Event:
        event = self._get_event(event_or_id)
        event.log = log_list
        self._set_event(event)
        return event



import os
from pathlib import Path
from tempfile import mkstemp


__all__ = [
    'clean_textfile',
    'make_tmptextfile',
    'delete_tmpfile',
    'executor',
]


def clean_textfile(text):
    text = "\n".join(line.strip() for line in text.split("\n"))
    return text


def make_tmptextfile(text, suffix, prefix=None, encoding='ascii'):
    text = clean_textfile(text)
    fd, filename = mkstemp(text=True, suffix=suffix, prefix=prefix)
    os.write(fd, bytes(text, encoding=encoding))
    return filename


def delete_tmpfile(filename):
    Path(filename).unlink(missing_ok=True)


def executor(client):
    d = {
        "user testuser 7f53cac4ffa877616b8472d3b33a44cbba1907ad  -\r\n": ["200 ok\r\n"],
        "user auth-failure 7f53cac4ffa877616b8472d3b33a44cbba1907ad  -\r\n": [
            "500 Authentication failure\r\n"
        ],
        "user illegal-first-response 85970fa23a2f5aa06c22b60f04013fe072319ebd  -\r\n": [
            "something-gurba-happened"
        ],
        "user no_login_response 87101b4944f4200fce90f519ddae5eacffeeadf6  -\r\n": [""],
        "caseids\r\n": [
            "304 list of active cases follows, terminated with '.'\r\n",
            "32802\r\n34978\r\n.\r\n",
        ],
        "getattrs 32802\r\n": [
            "303 simple attributes follow, terminated with '.'\r\n",
            "state: working\r\nrouter: uninett-gsw2\r\ntype: bgp\r\nopened: 1524465142\r\nremote-addr: 2001:700:0:4515::5:11\r\nid: 32802\r\npeer-uptime: 0\r\nupdated: 1533116751\r\npolladdr: 127.0.0.1\r\npriority: 100\r\nbgpOS: down\r\nbgpAS: halted\r\nremote-AS: 64666\r\nlastevent: peer is admin turned off\r\n.\r\n",
        ],
        "getattrs 34978\r\n": [
            "303 simple attributes follow, terminated with '.'\r\n",
            "router: bergen-sw1\r\nstate: working\r\ntype: alarm\r\nalarm-count: 1\r\nopened: 1529156235\r\nalarm-type: yellow\r\nid: 34978\r\nupdated: 1529156235\r\npolladdr: 127.0.0.1\r\npriority: 100\r\nlastevent: alarms went from 0 to 1\r\n.\r\n",
        ],
        "getattrs 40959\r\n": [
            "303 simple attributes follow, terminated with '.'\r\n",
            "state: open\r\nrouter: oslo-gw1\r\ntype: bgp\r\nopened: 1539480952\r\nremote-addr: 193.108.152.34\r\nid: 40959\r\npeer-uptime: 503\r\nupdated: 1539485757\r\npolladdr: 128.39.0.1\r\npriority: 500\r\nbgpOS: established\r\nremote-AS: 21357\r\nbgpAS: running\r\nlastevent: peer was reset (now up)\r\n.\r\n",
        ],
        "gethist 40959\r\n": [
            "301 history follows, terminated with '.'\r\n",
            "1539480952 state change embryonic -> open (monitor)\r\n1539509123 runarb\r\n Testmelding ifra pyRitz\r\n \r\n.\r\n",
        ],
        "getlog 40959\r\n": [
            "300 log follows, terminated with '.'\r\n",
            "1539480952 oslo-gw1 peer 193.108.152.34 AS 21357 was reset (now up)\r\n1539484557 oslo-gw1 peer 193.108.152.34 AS 21357 was reset (now up)\r\n1539485757 oslo-gw1 peer 193.108.152.34 AS 21357 was reset (now up)\r\n.\r\n",
        ],
        "addhist 40959  -\r\n": [
            "302 please provide new history entry, termiate with '.'\r\n"
        ],
        "Testmelding ifra pyRitz\r\n\r\n.\r\n": ["200 ok\r\n"],
        "setstate 40959 open\r\n": ["200 ok\r\n"],
        "setstate 40959 working\r\n": ["200 ok\r\n"],
        "setstate 40959 waiting\r\n": ["200 ok\r\n"],
        "setstate 40959 confirm-wait\r\n": ["200 ok\r\n"],
        "setstate 40959 ignored\r\n": ["200 ok\r\n"],
        "setstate 40959 closed\r\n": ["200 ok\r\n"],
        "setstate 40960 open\r\n": ["500 Cannot reopen closed event 40960\r\n"],
        "clearflap uninett-tor-sw3 707\r\n": ["200 ok\r\n"],
        "pollrtr uninett-tor-sw3": ["200 ok\r\n"],
        "pollintf uninett-tor-sw3 707": ["200 ok\r\n"],
        "ntie 909e90c2eda89a09819ee7fe9b3f67cadb31449f\r\n": ["200 ok\r\n"],
        "ntie xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\r\n": [
            "500 Could not find your notify socket\r\n"
        ],
    }

    client.send("200 2f88fe9d496b1c1a33a8d69f5c3ff7e8c34a1069 Hello, there\r\n")
    client.executor(d)



from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from zinolib.config.tcl import parse_tcl_config, normalize, parse

from ..utils import clean_textfile, make_tmptextfile, delete_tmpfile


RITZ_CONFIG = """
    set Secret 0123456789
    set User admin
    set Server example.org
    set Port 8001

    global Sortby
    set Sortby "upd-rev"

    set _Secret(dev-server) 0123456789
    set _User(dev-server) admin
    set _Server(dev-server) example.com
    set _Port(dev-server) 8001
"""


class ParseTclConfigTest(TestCase):
    def test_parse_tcl_config_missing_config_file(self):
        with self.assertRaises(FileNotFoundError):
            parse_tcl_config("cvfgdh vghj vbjhk")

    def test_parse_tcl_config_empty_config_file(self):
        filename = make_tmptextfile("", suffix=".tcl")
        tmp_directory = Path(filename).parent
        with patch("zinolib.config.utils.CONFIG_DIRECTORIES", [tmp_directory]):
            tcl_config_dict = parse_tcl_config(filename)
        delete_tmpfile(filename)
        self.assertEqual(tcl_config_dict, {})

    def test_parse_tcl_config_golden_path(self):
        filename = make_tmptextfile(RITZ_CONFIG, suffix=".tcl")
        tmp_directory = Path(filename).parent
        with patch("zinolib.config.utils.CONFIG_DIRECTORIES", [tmp_directory]):
            tcl_config_dict = parse_tcl_config(filename)
        delete_tmpfile(filename)
        expected = {
            "default": {
                "Port": "8001",
                "Secret": "0123456789",
                "Server": "example.org",
                "Sortby": '"upd-rev"',
                "User": "admin",
            },
            "dev-server": {
                "Port": "8001",
                "Secret": "0123456789",
                "Server": "example.com",
                "User": "admin",
            },
        }
        self.assertEqual(tcl_config_dict, expected)


class ParseNormalizeTest(TestCase):
    def test_normalize_empty_dict(self):
        expected = {"global_options": {}, "connections": {}}
        self.assertEqual(normalize({}), expected)

    def test_normalize_golden_path(self):
        tcl_config_dict = parse(clean_textfile(RITZ_CONFIG))
        expected = {
            "global_options": {"sort_by": '"upd-rev"'},
            "connections": {
                "default": {
                    "secret": "0123456789",
                    "username": "admin",
                    "server": "example.org",
                    "port": "8001",
                },
                "dev_server": {
                    "secret": "0123456789",
                    "username": "admin",
                    "server": "example.com",
                    "port": "8001",
                },
            },
        }
        self.assertEqual(normalize(tcl_config_dict), expected)



from os import environ
from pathlib import Path


DEFAULT_XDG_CONFIG_HOME = Path.home() / '.config'
XDG_CONFIG_HOME = environ.get('XDG_CONFIG_HOME', DEFAULT_XDG_CONFIG_HOME)
VISIBLE_LOCATIONS = [XDG_CONFIG_HOME, Path('/usr/local/etc'), Path('/etc')]
INVISIBLE_LOCATIONS = [Path.cwd(), Path.home()]
CONFIG_DIRECTORIES = INVISIBLE_LOCATIONS + VISIBLE_LOCATIONS


def make_filename_safe(filename):
    return Path(filename).name


def find_config_file(filename, directories=None):
    """
    Look for filename in ``directories`` in order

    The filename is validated. Any prefixed dot is removed, any path-magic (~,
    /, .. etc.) is stripped away.

    If the file isn't found in any of them, raise FileNotFoundError
    """
    if directories is None:
        directories = CONFIG_DIRECTORIES
    tried = []
    filename = filename.lstrip('.')
    filename = make_filename_safe(filename)
    for directory in directories:
        directory = Path(directory)
        if directory in INVISIBLE_LOCATIONS:
            used_filename = f'.{filename}'
        else:
            used_filename = filename
        path = directory / used_filename
        tried.append(path)
        if path.is_file():
            return path
    tried_paths = [str(path) for path in tried]
    raise FileNotFoundError(f"Looked for config in {tried_paths}, none found")



from pathlib import Path
import unittest
from unittest.mock import patch

from zinolib.config.toml import parse_toml_config

from ..utils import make_tmptextfile, delete_tmpfile


CONFIG = """
    [connections.default]
    server = "example.org"
    port = 8001
    username = "froofroo"
    password = "fghfgh"

    [options]
    autoremove = true
"""


class ParseTomlConfig(unittest.TestCase):

    def test_parse_toml_config_missing_config_file(self):
        with self.assertRaises(FileNotFoundError):
            parse_toml_config("hghjgjh jhgjhkjk ")

    def test_parse_toml_config_empty_config_file(self):
        filename = make_tmptextfile("", suffix=".toml")
        tmp_directory = Path(filename).parent
        with patch("zinolib.config.utils.CONFIG_DIRECTORIES", [tmp_directory]):
            toml_config_dict = parse_toml_config(filename)
        self.assertEqual(toml_config_dict, {})
        delete_tmpfile(filename)

    def test_parse_toml_config_golden_path(self):
        filename = make_tmptextfile(CONFIG, ".toml")
        tmp_directory = Path(filename).parent
        with patch("zinolib.config.utils.CONFIG_DIRECTORIES", [tmp_directory]):
            toml_config_dict = parse_toml_config(filename)
        expected = {
           "connections": {
               "default": {
                   "port": 8001,
                   "password": "fghfgh",
                   "server": "example.org",
                   "username": "froofroo",
               },
           },
           "options": {
               "autoremove": True,
           },
        }
        self.assertEqual(toml_config_dict, expected)
        delete_tmpfile(filename)



import unittest
from argparse import ArgumentParser


from zinolib.config.zino1 import ZinoV1Config, _parse_tcl


class ParseTclTest(unittest.TestCase):

    def test_parse_tcl_golden_path(self):
        section = "default"
        config_dict = {
            "default": {
                "Port": "8001",
                "Secret": "0123456789",
                "Server": "example.org",
                "Sortby": '"upd-rev"',
                "User": "admin",
            },
            "dev-server": {
                "Port": "8001",
                "Secret": "0123456789",
                "Server": "example.com",
                "User": "admin",
            },
        }
        expected_connection = {
            "port": 8001,
            "password": "0123456789",
            "server": "example.org",
            "username": "admin",
        }
        expected_option = {
            "sort_by": '"upd-rev"',
        }
        connection, options = _parse_tcl(config_dict, section)
        self.assertEqual(expected_connection, connection)
        self.assertEqual(expected_option, options)


class ZinoV1ConfigTest(unittest.TestCase):

    example_connection = {
        "port": "8001",
        "password": "0123456789",
        "server": "example.org",
        "username": "admin",
    }

    def manually_create_config(self, connection=None):
        connection = connection or self.example_connection
        _dict = {"connections": {"default": connection}}
        return ZinoV1Config.from_dict(_dict)

    def test_manually_create_config(self):
        config = self.manually_create_config()
        self.assertEqual(config.port, 8001)

    def test_set_userauth(self):
        config = self.manually_create_config()
        self.assertEqual(config.username, "admin")
        config.set_userauth('foo', 'barfybarf')
        self.assertEqual(config.username, "foo")

    def test_update_from_args(self):
        parser = ArgumentParser()
        parser.add_argument("password")
        parser.add_argument("unknown")
        args = parser.parse_args(["x", "y"])
        config = self.manually_create_config()
        self.assertEqual(config.password, "0123456789")
        config.update_from_args(args)
        self.assertEqual(config.password, "x")
        self.assertNotIn("unknown", vars(config))



from typing import ClassVar

from pydantic import BaseModel

from . import tcl, toml
from .models import UserConfig, OptionalUserConfig, ServerV1Config, Options


def _parse_tcl(config_dict, section):
    fixed_dict = tcl.normalize(config_dict)
    connection = fixed_dict["connections"][section]
    options = fixed_dict["global_options"]
    connection['password'] = connection.pop("secret")
    connection['port'] = int(connection.pop("port"))
    return connection, options


class ZinoV1Config(ServerV1Config, Options):
    """
    How to use::

    Given a legacy tcl config file stored on disk, containing at minimum server,
    username and secret::

        > config = ZinoV1Config.from_tcl()

    With a toml-file stored on disk, which needs only a server::

        > config = ZinoV1Config.from_toml()

    Explicitly set the user and Zino1 secret::

        > config.set_userauth(username, secret)

    Read some command-line arguments via argparse.ArgumentParser and update the
    config::

        > config.update_from_args(args)
    """
    DEFAULT_SECTION: ClassVar = "default"

    @classmethod
    def get_legacy_class(cls):
        return type("ZinoV1LegacyConfig", (UserConfig, cls), dict())

    @classmethod
    def get_class(cls):
        return type("ZinoV1Config", (OptionalUserConfig, cls), dict())

    @classmethod
    def from_dict(cls, config_dict, section=DEFAULT_SECTION):
        connection = config_dict["connections"][section]
        options = config_dict.get("options", {})
        classobj = cls.get_class()
        return classobj(**connection, **options)

    @classmethod
    def from_tcl(cls, filename=None, section=DEFAULT_SECTION):
        config_dict = tcl.parse_tcl_config(filename)
        connection, options = _parse_tcl(config_dict, section)
        classobj = cls.get_legacy_class()
        return classobj(**connection, **options)

    @classmethod
    def from_toml(cls, filename=None, section=DEFAULT_SECTION):
        config_dict = toml.parse_toml_config(filename)
        return cls.from_dict(config_dict)

    def set_userauth(self, username, password):
        self.username = username
        self.password = password

    def update_from_args(self, args):
        """
        Assumes argparse-style args namespace object

        arg-names not found in the config-object are ignored.
        """
        for arg in vars(args):
            value = getattr(args, arg, None)
            if arg in vars(self):
                setattr(self, arg, value)



"""
Expected format of toml-file::

    [connections.default]
    server = "SERVERNAME"
    port = PORT
    username = "SOMEUSERNAME"  # [A-Za-z][A-Za-z0-9]+
    password = "PASSWORD"

    [options]
    timeout = 30
    autoremove = false

This is parsed into a dict of the format::

    {
        "connections": {
            "default": {
                "server": "SERVERNAME",
                "port: PORT,
                "username": "SOMEUSERNAME"
                "password": "PASSWORD"
            },
        },
        "options": {
            "timeout": 30,
            "autoremove": False,
        },
    }
"""

from zinolib.compat import tomlload
from zinolib.config.utils import find_config_file


def parse_toml_config(filename=None):
    full_filename = find_config_file(filename)
    if full_filename:
        with open(full_filename, "rb") as TF:
            toml = tomlload(TF)
            return toml



"""
Get a live Zino 1 session to use::

    > from zinolib.config.zino1 import ZinoV1Config
    > from zinolib.controllers.zino1 import Zino1EventManager
    > config = ZinoV1Config.from_tcl('.ritz.tcl', 'default')
    > event_manager = Zino1EventManager.configure(config)
    > event_manager.connect()

The configuration can also be stored in a toml-file::

    > config = ZinoV1Config.from_toml('.ritz.toml', 'default')

.. or in a dict with the format::

    > config_dict = {"connections": "default": { "server": ACTUAL CONFIG .. }}
    > config = ZinoV1Config.from_dict(config_dict, 'default')

Do not initialize ZinoV1Config directly, avoid ``ZinoV1Config(**dict)``.

Authenticate using a username and password from the config-file::

    > event_manager.authenticate()

Explicitly autenticate with a username and password::

    > event_manager.authenticate(username, password)

To get a list of currently available events::

    > event_manager.get_events()

The events are then available as::

    > event_manager.events

This is a dictionary of event_id, event object pairs.

To get a set of removed event ids::

    > event_manager.removed_ids

For updates, either regularly use ``get_events()`` or utilize the UpdateHandler::

    > updater = UpdateHandler(event_manager)
    > updater.connect()
    > updated = updater.get_event_update()

The updater is unique per authenticated user and is only available after login.

``updater.get_event_update()`` updates ``event_manager.events`` and
``event_manager.removed_ids`` and returns the id of a changed event on any
change, falsey otherwise. Check the id against the removed_id's set to see if
it has been removed from the server.

To get history for a specific event::

    > history_list = event_manager.get_history_for_id(INT)

To get the log for a specific event::

    > log_list = event_manager.get_log_for_id(INT)

History can be stored on the correct event with one of::

    > event_manager.set_history_for_event(event, history_list)
    > event_manager.set_history_for_event(INT, history_list)

Log can be stored on the correct event with one of::

    > event_manager.set_log_for_event(event, log_list)
    > event_manager.set_log_for_event(INT, log_list)

Both return the changed event.

The adapters are not meant to be used directly.
"""

from datetime import datetime, timezone
from typing import Dict, Iterable, List, TypedDict, Optional, Set
import logging

from .base import EventManager, EventOrId
from ..compat import StrEnum
from ..event_types import EventType, Event, HistoryEntry, LogEntry, AdmState
from ..ritz import ZinoError, ProtocolError, ritz, notifier, NotConnectedError
from ..utils import log_exception_with_params


__all__ = [
    'Zino1EventManager',
]


HistoryDict = TypedDict(
    "HistoryDict",
    {"date": datetime, "user": str, "log": str},
    total=False,
)
LogDict = TypedDict(
    "LogDict",
    {"log": str, "date": datetime},
    total=False,
)


DEFAULT_TIMEOUT = 30
LOG = logging.getLogger(__name__)


class Zino1Error(ZinoError):
    pass


class RetryError(Zino1Error):
    pass


class EventClosedError(Zino1Error):
    pass


class LostConnectionError(NotConnectedError):
    pass


def convert_timestamp(timestamp: int) -> datetime:
    return datetime.fromtimestamp(timestamp, timezone.utc)


class UpdateHandler:
    class UpdateType(StrEnum):
        STATE = "state"
        ATTR = "attr"
        HISTORY = "history"
        LOG = "log"
        SCAVENGED = "scavenged"

    class UpdateError(Zino1Error):
        pass

    class SocketError(UpdateError):
        pass

    def __init__(self, manager, autoremove=False):
        self._connected = False
        if not manager.is_authenticated:
            msg = "Cannot initiate update handler, not authenticated"
            raise self.UpdateError(msg)
        self.manager = manager
        self.events = manager.events
        self.autoremove = autoremove

    def connect(self):
        if not self.manager.session.push:
            self.manager.connect_push_channel()
        self.check_connection()

    def check_connection(self):
        self.manager._verify_session()
        if self.manager.session.push._sock.fileno() >= 0:
            self._connected = True
            return True
        self._connected = False
        raise self.SocketError("Push socket reports failure, fileno = -1")

    def get_event_update(self):
        """
        Fetches one update for a single event and runs the appropriate handler

        Attributes on the update object:

        id: event id
        type: update type, triggers the correct handler
        info: a type-specific string with the actual change

        Run in a loop/every N seconds for a lightweight way to update the event
        list
        """
        self.check_connection()
        update = self.manager.session.push.poll()
        if not update:
            return False
        return self.handle_event_update(update)

    def update(self, event_id: int):
        "Refresh an event from the server, refreshing everything"
        event = self.manager.get_updated_event_for_id(event_id)
        self.manager._set_event(event)
        LOG.debug("Updated event #%i", event_id)

    def remove(self, event_id: int):
        "Remove an event from our local copy of the events list"
        self.manager.remove_event(event_id)
        LOG.debug("Removed event #%i", event_id)

    def handle_event_update(self, update):
        """Call the right handle on the update object depending on type

        If the update is about a locally unknown id and the update type is not
        UpdateType.STATE it is a new, incomplete event so we return nothing.

        Otherwise call the right handler on the update data.
        """
        if update.id not in self.events and update.type != self.UpdateType.STATE:
            # unknown event that don't have a state (yet), wait for new update
            return None
        if update.type in tuple(self.UpdateType):
            method = getattr(self, f"cmd_{update.type}")
            return method(update)
        return self.fallback(update)

    def cmd_state(self, update):
        """State has been changed

        Removes a now closed state if the setting "autoremove" is True,
        otherwise refreshes the event from the server.
        """
        states = update.info.split(" ")
        if states[1] == "closed" and self.autoremove:
            LOG.debug('Autoremoving "%s"', update.id)
            self.remove(update.id)
        else:
            self.update(update.id)
        return update.id

    def cmd_attr(self, update):
        """Attributes has been changed

        Refresh the event from the server.
        """
        self.update(update.id)
        return update.id

    cmd_history = cmd_attr
    cmd_log = cmd_attr

    def cmd_scavenged(self, update):
        """The event has been removed from the server

        Remove it from our local copy of the events list.
        """
        self.remove(update.id)
        return update.id

    def fallback(self, update):
        """There's an unknown update type

        Log it and do nothing.
        """
        LOG.warning('Unknown update type: "%s" for id %s' % (update.type, update.id))
        return False


class SessionAdapter:

    class _Session:
        push = None
        request = None

    @classmethod
    def create_session(cls, config):
        session = cls._setup_request(cls._Session, config)
        return session

    @staticmethod
    def _setup_request(session, config):
        if not session.request:
            session.request = ritz(
                config.server,
                username=config.username,
                password=config.password,
                timeout=config.timeout,
            )
        return session

    @classmethod
    def connect_session(cls, session):
        session.request.connect()
        return cls.connect_push_channel(session)

    @staticmethod
    def connect_push_channel(session):
        if session.request.connected and session.request.authenticated:
            session.push = notifier(session.request)
            session.push.connect()  # ntie
        return session

    @staticmethod
    def authenticate(session, username=None, password=None):
        username = username if username else session.request.username
        password = password if password else session.request.password
        if not username and password:
            raise ValueError("Both username and password must be set and truthy")
        session.request.authenticate(username, password)
        return session

    @staticmethod
    def close_push_channel(session):
        if hasattr(session.push, '_sock'):
            session.push._sock.close()
        session.push = None

    @classmethod
    def close_session(cls, session):
        cls.close_push_channel(session)
        session.request.close()
        session.request = None
        return None


class EventAdapter:
    FIELD_MAP = {
        "state": "adm_state",
        "bfdAddr": "bfd_addr",
        "bfdDiscr": "bfd_discr",
        "bfdState": "bfd_state",
        "bfdIx": "bfd_ix",
        # "Neigh-rDNS": "Neigh_rDNS",
        "bgpAS": "bgp_AS",
        "bgpOS": "bgp_OS",
        "ifindex": "if_index",
        "portstate": "port_state",
    }
    FIELD_VALUE_CONVERTER_MAP = {
        'ac_down': int,
    }

    @staticmethod
    def get_attrlist(request, event_id: int):
        return request.get_raw_attributes(event_id)

    @staticmethod
    def validate_raw_attrlist(attrlist):
        for item in attrlist:
            if ':' not in item:
                LOG.error('"getattrs" is spewing garbage: %s', attrlist)
                return False
        return True

    @classmethod
    @log_exception_with_params(LOG)
    def attrlist_to_attrdict(cls, attrlist: Iterable[str]) -> Dict[str, str]:
        """Translate a wire protocol dump of a single event

        The dump is a list of lines of the format:

            attr:value
        """
        event_dict = {}
        for item in attrlist:
            k, v = item.split(":", 1)
            if '-' in k:
                k = '_'.join(k.split('-'))
            k = cls.FIELD_MAP.get(k, k)
            event_dict[k.strip()] = v.strip()
        return event_dict

    @classmethod
    def convert_values(cls, attrdict):
        "Convert some values that pydantic handles clumsily"
        for field_name, function in cls.FIELD_VALUE_CONVERTER_MAP.items():
            if field_name in attrdict:
                value = function(attrdict[field_name])
                attrdict[field_name] = value
        return attrdict

    @staticmethod
    def set_admin_state(request, event: EventType, state: AdmState) -> bool:
        return request.set_state(event.id, state.value)

    @staticmethod
    def get_event_ids(request):
        try:
            return request.get_caseids()
        except ProtocolError as e:
            raise RetryError('Zino 1 failed to send a correct response header, retry') from e
        except BrokenPipeError as e:
            raise LostConnectionError('Lost connection to Zino 1 server') from e

    @staticmethod
    def poll(request, event: EventType) -> bool:
        if event.type == Event.Type.PORTSTATE:
            return request.poll_interface(event.router, event.if_index)
        else:
            return request.poll_router(event.router)


class HistoryAdapter:
    SYSTEM_USER = "monitor"

    @staticmethod
    def get_history(request, event_id: int):
        return request.get_raw_history(event_id).data

    @classmethod
    @log_exception_with_params(LOG)
    def parse_response(cls, history_data: Iterable[str]) -> list[HistoryDict]:
        """
        Input:

        [
            '1678273372 state change embryonic -> open (monitor)',
            '1678276375 someuser',
            ' manually recorded history message ',
            ' ',
            '1678276378 state change open -> waiting (someuser)',
            '1680265996 someotheruser',
            ' other manually recorded history message ',
            ' ',
            '1680266003 state change waiting -> working (someotheruser)']

        Output:
        [
            {
                "date": datetime,
                "user": user,
                "log": log,
            },
            ..
        ]
        """
        history_list: List[HistoryDict] = []
        for row in history_data:
            if row == " ":  # end of history
                continue
            if row.startswith(" "):  # history
                history_list[-1]["log"] += row.strip() + ' '
            else:
                timestamp, body = row.split(" ", 1)
                dt = convert_timestamp(int(timestamp))
                entry: HistoryDict = {"date": dt}
                if " " in body:  # server generated
                    entry["user"] = cls.SYSTEM_USER
                    entry["log"] = body
                else:
                    entry["user"] = body
                    entry["log"] = ""
                history_list.append(entry)
        for entry in history_list:
            entry["log"] = entry["log"].strip()

        return history_list

    @classmethod
    def add(cls, request, message: str, event: EventType) -> Optional[EventType]:
        success = request.add_history(event.id, message)
        if success:
            # fetch history
            raw_history = cls.get_history(request, event.id)
            parsed_history = cls.parse_response(raw_history)
            new_history = HistoryEntry.create_list(parsed_history)
            if new_history != event.history:
                event.history = new_history
            return event
        return None


class LogAdapter:
    @staticmethod
    def get_log(request, event_id: int) -> list[str]:
        return request.get_raw_log(event_id).data

    @staticmethod
    @log_exception_with_params(LOG)
    def parse_response(log_data: Iterable[str]) -> list[LogDict]:
        """
        Input:
        [
            '1683159556 some log message',
            '1683218672 some other log message',
        ]

        Output:
        [
            {
                "date": datetime,
                "log": log
            },
            ..
        ]
        """
        log_list: List[LogDict] = []
        for row in log_data:
            timestamp, log = row.split(" ", 1)
            try:
                timestamp = int(timestamp)
            except ValueError as e:
                raise RetryError('Zino 1 did not send a log, retry') from e
            dt = convert_timestamp(timestamp)
            log_list.append({"date": dt, "log": log})
        return log_list


class Zino1EventManager(EventManager):
    # Easily replaced in order to ease testing
    _session_adapter = SessionAdapter
    _event_adapter = EventAdapter
    _history_adapter = HistoryAdapter
    _log_adapter = LogAdapter
    removed_ids: Set[int] = set()
    config = None

    @property
    def is_authenticated(self):
        session_ok = self._verify_session(quiet=True)
        return self.session.request.authenticated if session_ok else False

    @property
    def is_connected(self):
        session_ok = self._verify_session(quiet=True)
        return self.session.request.connected if session_ok else False

    def rename_exception(self, function, *args):
        "Replace the original exception with our own"
        try:
            return function(*args)
        except (NotConnectedError, LostConnectionError) as e:
            raise
        except ZinoError as e:
            raise self.ManagerException(e) from e

    def _verify_session(self, quiet=False):
        if not getattr(self.session, 'request', None):
            if quiet:
                return False
            raise NotConnectedError("The request socket have not been set up correctly. Reconnect necessary.")
        if not self.session.request.connected:
            if quiet:
                return False
            raise NotConnectedError("Authentication necessary")
        return True

    @classmethod
    def configure(cls, config):
        session = cls._session_adapter.create_session(config)
        classobj = cls(session)
        classobj.config = config
        return classobj

    def connect(self):
        if not self._verify_session(quiet=True):
            self.session =  self._session_adapter.create_session(self.config)
        self.session = self._session_adapter.connect_session(self.session)

    def connect_push_channel(self):
        self.session = self._session_adapter.connect_push_channel(self.session)

    def authenticate(self, username=None, password=None):
        try:
            self.session = self._session_adapter.authenticate(self.session, username, password)
        except (ZinoError, ValueError) as e:
            raise self.ManagerException(e)

    def disconnect(self):
        session_ok = self._verify_session(quiet=True)
        if session_ok:
            self.session = self._session_adapter.close_session(self.session)
        else:
            self._session_adapter.close_push_channel(self.session)

    def clear_flapping(self, event_or_id: EventOrId):
        """Clear flapping state of a PortStateEvent

        Usage:
            c = ritz_session.case(123)
            c.clear_flapping()
        """
        event = self._get_event(event_or_id)
        if event.type == Event.Type.PORTSTATE:
            return self.session.request.clear_flapping(event.router, event.if_index)
        return None

    def poll(self, event_or_id: EventOrId):
        """Ask the server to refresh data for the event

        If there are any changes they will be available through the update
        handler in a bit
        """
        event = self._get_event(event_or_id)
        return self._event_adapter.poll(self.session.request, event)

    def get_events(self):
        self._verify_session()
        for event_id in self._event_adapter.get_event_ids(self.session.request):
            try:
                event = self.create_event_from_id(event_id)
            except self.ManagerException:
                self.remove_event(event_id)
                continue
            self.events[event_id] = event

    def test_connection(self):
        """Try fetching info about a non-existing event

        If the connection is up, we get a ProtocolError due to the
        non-existent event. Do nothing.

        If the connection is down we will pass on a TimeoutError.
        """
        try:
            self._event_adapter.get_attrlist(self.session.request, 0)
        except ProtocolError:
            pass

    def create_event_from_id(self, event_id: int):
        self._verify_session()
        attrlist = self.rename_exception(self._event_adapter.get_attrlist, self.session.request, event_id)
        if not self._event_adapter.validate_raw_attrlist(attrlist):
            raise RetryError('Zino 1 did not send event attributes, retry')
        attrdict = self._event_adapter.attrlist_to_attrdict(attrlist)
        attrdict = self._event_adapter.convert_values(attrdict)
        return Event.create(attrdict)

    def get_updated_event_for_id(self, event_id):
        event = self.create_event_from_id(event_id)
        history_list = self.get_history_for_id(event.id)
        self.set_history_for_event(event, history_list)
        log_list = self.get_log_for_id(event.id)
        self.set_log_for_event(event, log_list)
        return event

    def change_admin_state_for_id(self, event_id, admin_state: AdmState) -> Optional[Event]:
        self._verify_session()
        event = self._get_event(event_id)
        try:
            success = self._event_adapter.set_admin_state(self.session.request, event, admin_state)
        except ValueError as e:
            if 'reopen' in str(e):
                raise EventClosedError("Cannot set state on closed event")
            else:
                raise
        if success:
            event = self.get_updated_event_for_id(event_id)
            self._set_event(event)
            return event
        return None

    def get_history_for_id(self, event_id: int) -> list[HistoryEntry]:
        self._verify_session()
        raw_history = self.rename_exception(self._history_adapter.get_history, self.session.request, event_id)
        parsed_history = self._history_adapter.parse_response(raw_history)
        return HistoryEntry.create_list(parsed_history)

    def add_history_entry_for_id(self, event_id: int, message) -> Optional[Event]:
        self._verify_session()
        event = self._get_event(event_id)
        success = self._history_adapter.add(self.session.request, message, event)
        if success:
            event = self.get_updated_event_for_id(event_id)
            self._set_event(event)
            return event
        return None

    def get_log_for_id(self, event_id: int) -> list[LogEntry]:
        self._verify_session()
        raw_log = self.rename_exception(self._log_adapter.get_log, self.session.request, event_id)
        parsed_log = self._log_adapter.parse_response(raw_log)
        return LogEntry.create_list(parsed_log)



# mypy: ignore-errors
try:
    from enum import StrEnum
except ImportError:
    # < Python 3.11
    from enum import Enum

    class StrEnum(str, Enum):
        __str__ = str.__str__

        def _generate_next_value_(name, start, count, last_values):
            """
            Return the lower-cased version of the member name.
            """
            return name.lower()


try:
    from tomllib import load as tomlload
except ImportError:
    try:
        from tomli import load as tomlload
    except ImportError:
        def tomlload(*args, **kwargs):
            raise ImportError("TOML library not available, TOML configuration impossible")


__all__ = ["StrEnum", "tomlload"]



import unittest
from datetime import datetime, timedelta, timezone

from zinolib.event_types import AdmState, Event, HistoryEntry, LogEntry
from zinolib.controllers.zino1 import EventAdapter, HistoryAdapter, LogAdapter, SessionAdapter, Zino1EventManager, UpdateHandler
from zinolib.controllers.zino1 import RetryError, NotConnectedError
from zinolib.ritz import NotifierResponse

raw_event_id = 139110
raw_attrlist = [
    'state: ignored',
    'router: uninett-tor-sw4',
    'descr: local MC-LAG, tor-sw4.har-04.p.uninett.no',
    'type: portstate',
    'opened: 1677714463',
    'lasttrans: 1686584585',
    'id: 139110',
    'port: ae24',
    'flaps: 1',
    'updated: 1686584585',
    'ac-down: 352952',
    'priority: 100',
    'polladdr: 158.38.129.42',
    'portstate: up',
    'ifindex: 654',
]
raw_history = [
    '1678273372 state change embryonic -> open (monitor)',
    '1678276375 someuser',
    ' manually recorded history message ',
    ' ',
    '1678276378 state change open -> waiting (someuser)',
    '1680265996 someotheruser',
    ' other manually recorded history message ',
    ' ',
    '1680266003 state change waiting -> working (someotheruser)'
]
raw_log = [
    '1683159556 some log message',
    '1683218672 some other log message',
]


class FakeEventAdapter:
    @staticmethod
    def get_attrlist(request, event_id: int):
        return raw_attrlist.copy()

    @classmethod
    def attrlist_to_attrdict(cls, attrlist):
        return EventAdapter.attrlist_to_attrdict(attrlist)

    @staticmethod
    def validate_raw_attrlist(attrlist):
        return True

    @classmethod
    def convert_values(cls, attrdict):
        return EventAdapter.convert_values(attrdict)

    @staticmethod
    def get_event_ids(request):
        return [raw_event_id]


class FakeHistoryAdapter(HistoryAdapter):
    @staticmethod
    def get_history(request, event_id: int):
        return raw_history.copy()


class FakeLogAdapter(LogAdapter):
    @staticmethod
    def get_log(request, event_id: int):
        return raw_log.copy()


class FakeSessionAdapter(SessionAdapter):

    @classmethod
    def _setup_config(cls, config):
        pass

    @staticmethod
    def _setup_request(session, config):
        class FakeSession:
            authenticated = True
            connected = True

        session.request = FakeSession()  # needs to be truthy
        session.push = True  # needs to be truthy
        return session


class FakeZino1EventManager(Zino1EventManager):
    _event_adapter = FakeEventAdapter
    _history_adapter = FakeHistoryAdapter
    _log_adapter = FakeLogAdapter
    _session_adapter = FakeSessionAdapter

    def __init__(self, session=None):
        super().__init__(session)


class Zino1EventManagerTest(unittest.TestCase):

    def init_manager(self):
        zino1 = FakeZino1EventManager.configure(None)
        return zino1

    def test_verify_session_raises_notconnectederror_on_incorrect_manager(self):
        zino1 = FakeZino1EventManager()
        with self.assertRaises(NotConnectedError) as e:
            zino1._verify_session()
            self.assertIn("The request socket have not been set up correctly", e)

    def test_verify_session_raises_notconnectederror_if_not_connected(self):
        zino1 = FakeZino1EventManager.configure(None)
        orig = zino1.session.request.connected
        zino1.session.request.connected = False
        try:
            with self.assertRaises(NotConnectedError) as e:
                zino1._verify_session()
                self.assertIn("Authentication necessary", e)
        finally:
            zino1.session.request.connected = orig

    def test_verify_session_returns_False_if_quieted_on_incorrect_manager(self):
        zino1 = FakeZino1EventManager()
        result = zino1._verify_session(quiet=True)
        self.assertEqual(result, False)

    def test_create_event_from_id_receiving_garbage_admstate_is_safely_handled(self):
        global raw_attrlist
        zino1 = self.init_manager()
        good_attrlist = raw_attrlist[:] # copy
        try:
            raw_attrlist[0] = "state: garbage admstate"
            event = zino1.create_event_from_id(139110)
            self.assertEqual(event.adm_state, AdmState.UNKNOWN)
        finally:
            # reset to known good attrlist for other tests
            raw_attrlist = good_attrlist

    def test_create_event_from_id_may_get_garabage_data(self):
        def falsey(_):
            return False

        zino1 = self.init_manager()
        old = zino1._event_adapter.validate_raw_attrlist
        zino1._event_adapter.validate_raw_attrlist = staticmethod(falsey)
        with self.assertRaises(RetryError):
            zino1.create_event_from_id(139110)
        zino1._event_adapter.validate_raw_attrlist = old

    def test_get_events(self):
        zino1 = self.init_manager()
        self.assertEqual(len(zino1.events), 0)
        zino1.get_events()
        self.assertEqual(len(zino1.events), 1)
        self.assertIn(raw_event_id, zino1.events)
        self.assertEqual(zino1.events[raw_event_id].id, raw_event_id)

    def test_get_history_for_id(self):
        zino1 = self.init_manager()
        history_list = zino1.get_history_for_id(4567)
        expected_history_list = [
            HistoryEntry(
                log='state change embryonic -> open (monitor)',
                date=datetime(2023, 3, 8, 11, 2, 52, tzinfo=timezone.utc),
                user='monitor'),
            HistoryEntry(
                log='manually recorded history message',
                date=datetime(2023, 3, 8, 11, 52, 55, tzinfo=timezone.utc),
                user='someuser'),
            HistoryEntry(
                log='state change open -> waiting (someuser)',
                date=datetime(2023, 3, 8, 11, 52, 58, tzinfo=timezone.utc),
                user='monitor'),
            HistoryEntry(
                log='other manually recorded history message',
                date=datetime(2023, 3, 31, 12, 33, 16, tzinfo=timezone.utc),
                user='someotheruser'),
            HistoryEntry(
                log='state change waiting -> working (someotheruser)',
                date=datetime(2023, 3, 31, 12, 33, 23, tzinfo=timezone.utc),
                user='monitor')
        ]
        self.assertEqual(history_list, expected_history_list)

    def test_get_log_for_id(self):
        zino1 = self.init_manager()
        log_list = zino1.get_log_for_id(4567)
        expected_log_list = [
            LogEntry(
                date=datetime(2023, 5, 4, 0, 19, 16, tzinfo=timezone.utc),
                log='some log message'),
            LogEntry(
                date=datetime(2023, 5, 4, 16, 44, 32, tzinfo=timezone.utc),
                log='some other log message')
        ]
        self.assertEqual(log_list, expected_log_list)


class UpdateHandlerTest(unittest.TestCase):

    def init_manager(self):
        zino1 = FakeZino1EventManager.configure(None)
        return zino1

    def test_cmd_scavenged(self):
        zino1 = self.init_manager()
        zino1.get_events()
        self.assertIn(raw_event_id, zino1.events)
        self.assertNotIn(raw_event_id, zino1.removed_ids)
        updates = UpdateHandler(zino1)
        update = NotifierResponse(raw_event_id, "","")
        ok = updates.cmd_scavenged(update)
        self.assertTrue(ok)
        self.assertNotIn(raw_event_id, zino1.events)
        self.assertIn(raw_event_id, zino1.removed_ids)

    def test_cmd_attr(self):
        zino1 = self.init_manager()
        zino1.get_events()
        old_events = zino1.events.copy()
        old_events[raw_event_id].priority = 500
        updates = UpdateHandler(zino1)
        update = NotifierResponse(raw_event_id, "","")
        ok = updates.cmd_attr(update)
        self.assertTrue(ok)
        self.assertNotEqual(zino1.events[raw_event_id].priority, old_events[raw_event_id].priority)

    def test_cmd_state_is_closed_and_autoremove_is_on(self):
        zino1 = self.init_manager()
        zino1.get_events()
        self.assertNotIn(raw_event_id, zino1.removed_ids)
        self.assertIn(raw_event_id, zino1.events)
        updates = UpdateHandler(zino1, autoremove=True)
        update = NotifierResponse(raw_event_id, "", "X closed")
        ok = updates.cmd_state(update)
        self.assertTrue(ok)
        self.assertIn(raw_event_id, zino1.removed_ids)
        self.assertNotIn(raw_event_id, zino1.events)

    def test_cmd_state_is_closed_and_autoremove_is_off(self):
        zino1 = self.init_manager()
        zino1.get_events()
        old_events = zino1.events.copy()
        old_events[raw_event_id].priority = 500
        updates = UpdateHandler(zino1, autoremove=False)
        update = NotifierResponse(raw_event_id, "","X closed")
        ok = updates.cmd_state(update)
        self.assertTrue(ok)
        self.assertNotEqual(zino1.events[raw_event_id].priority, old_events[raw_event_id].priority)

    def test_cmd_state_is_not_closed(self):
        zino1 = self.init_manager()
        zino1.get_events()
        old_events = zino1.events.copy()
        old_events[raw_event_id].priority = 500
        updates = UpdateHandler(zino1, autoremove=False)
        update = NotifierResponse(raw_event_id, "","x butterfly")
        ok = updates.cmd_state(update)
        self.assertTrue(ok)
        self.assertNotEqual(zino1.events[raw_event_id].priority, old_events[raw_event_id].priority)

    def test_fallback(self):
        zino1 = self.init_manager()
        updates = UpdateHandler(zino1)
        update = NotifierResponse(raw_event_id, "", "")
        with self.assertLogs('zinolib.controllers.zino1', level='WARNING') as cm:
            self.assertFalse(updates.fallback(update))

    def test_handle_new_stateless_event_is_very_special(self):
        zino1 = self.init_manager()
        updates = UpdateHandler(zino1)
        update = NotifierResponse(1337, "", "")
        result = updates.handle_event_update(update)
        self.assertEqual(result, None)

    def test_handle_known_type(self):
        zino1 = self.init_manager()
        zino1.get_events()
        old_events = zino1.events.copy()
        old_events[raw_event_id].priority = 500
        updates = UpdateHandler(zino1)
        update = NotifierResponse(raw_event_id, updates.UpdateType.LOG, "")
        ok = updates.handle_event_update(update)  # will refetch events
        self.assertTrue(ok)
        self.assertNotEqual(zino1.events[raw_event_id].priority, old_events[raw_event_id].priority)

    def test_handle_unknown_type(self):
        zino1 = self.init_manager()
        zino1.get_events()
        updates = UpdateHandler(zino1)
        update = NotifierResponse(raw_event_id, "trout", "")
        with self.assertLogs('zinolib.controllers.zino1', level='WARNING'):
            ok = updates.handle_event_update(update)  # will run fallback
            self.assertFalse(ok)



import logging
from datetime import datetime, timedelta, timezone
from typing import Optional, ClassVar, List, TypeVar, Union, Dict, Generic, Set
from typing_extensions import Annotated

from pydantic import ConfigDict, IPvAnyAddress, ValidationError, ValidationInfo
from pydantic import BaseModel, computed_field, field_validator
from pydantic.functional_validators import BeforeValidator

from .compat import StrEnum
from .utils import log_exception_with_params


LOG = logging.getLogger(__name__)


def unknown(v: Optional[str]) -> Optional[str]:
    if v:
        if 'unknown' in v:
            return None
        return v
    return None


# Must be after "unknown"
OptionalIpAnyAddress = Annotated[Optional[IPvAnyAddress], BeforeValidator(unknown)]


def utcnow():
    return datetime.now(timezone.utc)


class AdmState(StrEnum):
    """

    Historic states:

    * "active" was used twice in 1998
    """
    # ACTIVE = "active"

    CLOSED = "closed"
    CONFIRM_WAIT = "confirm-wait"
    IGNORED = "ignored"
    OPEN = "open"
    WAITING = "waiting"
    WORKING = "working"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value):
        # Python 3.12: "if not value in cls:"
        if not value in cls._value2member_map_:
            return cls.UNKNOWN


class FlapState(StrEnum):
    FLAPPING = 'flapping'
    STABLE = 'stable'


class BFDState(StrEnum):
    ADMIN_DOWN = "adminDown"
    DOWN = "down"
    INIT = "init"
    UP = "up"


class ReachabilityState(StrEnum):
    REACHABLE = "reachable"
    NORESPONSE = "no-response"


class PortState(StrEnum):
    """
    Historic states:

    * "testing" was used six times from 2000 to 2001
    * "admin0" was used ten times from 2000 to 2002
    * "flapping" was used 22 times on 2016-07-13T09:55 UTC
    * "notPresent" was used 161 times during 2003
    * "5" was used 389 times from 1998 to 2000
    * "dormant" was used 813 times from 2000 to 2006
    """
    # TESTING = "testing"
    # ADMIN_0 = "admin0"
    # FLAPPING = "flapping"
    # NOT_PRESENT = "notPresent"
    # FIVE = "5"
    # DORMANT = "dormant"

    ADMIN_DOWN = "adminDown"
    DOWN = "down"
    LOWER_LAYER_DOWN = "lowerLayerDown"
    UP = "up"


class LogEntry(BaseModel):
    date: datetime  # epoch
    log: str

    @classmethod
    def create_list(cls, raw_log_list):
        log_list = []
        for entry in raw_log_list:
            obj = cls(**entry)
            log_list.append(obj)
        return log_list


class HistoryEntry(BaseModel):
    log: str
    date: datetime  # epoch
    user: str

    @classmethod
    def create_list(cls, raw_history_list):
        history_list = []
        for entry in raw_history_list:
            obj = cls(**entry)
            history_list.append(obj)
        return history_list


class Event(BaseModel):
    """
    Base-class with common fields for all events.

    Here there be dragons!
    The following cannot be set here, only per subclass::

        port: str = ""
        description: str = ""
        op_state: str = ""

    For that reason, never ``Event()``, always ``Event.create()``!
    """
    class Type(StrEnum):
        ALARM = "alarm"
        BFD = "bfd"
        BGP = "bgp"
        PORTSTATE = "portstate"
        REACHABILITY = "reachability"

    id: int
    type: Union[Type, str]
    adm_state: AdmState
    router: str
    opened: datetime  # epoch

    lastevent: Optional[str] = ''
    lasttrans: Optional[datetime] = None  # epoch
    updated: Optional[datetime] = None # epoch
    polladdr: Optional[IPvAnyAddress] = None
    priority: int = 100

    log: List[LogEntry] = []
    history: List[HistoryEntry] = []

    EXTRAS: ClassVar[Set[str]] = set()
    SUBTYPES: ClassVar[dict] = {}

    model_config = ConfigDict(validate_default=True)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.SUBTYPES[cls.type.value] = cls

    @classmethod
    @log_exception_with_params(LOG)
    def create(cls, attrdict):
        event_type_string = attrdict["type"]
        subtype = Event.SUBTYPES[event_type_string]
        eventobj = subtype(**attrdict)
        return eventobj

    def is_down(self):
        raise NotImplementedError("Abstract method was called, can not calculate is_down status without event type.")


EventType = TypeVar('EventType', bound=Event)
EventOrId = Union[EventType, int]


class AlarmEvent(Event):
    type: str = Event.Type.ALARM
    alarm_count: int
    alarm_type: str
    port: str = ""

    @computed_field  # type: ignore
    @property
    def op_state(self) -> str:
        return f"ALRM  {self.alarm_type}"

    @computed_field  # type: ignore
    @property
    def description(self) -> Optional[str]:
        return self.lastevent

    def is_down(self) -> bool:
        return self.alarm_count > 0


class BFDEvent(Event):
    type: str = Event.Type.BFD
    bfd_addr: OptionalIpAnyAddress = None
    bfd_discr: Optional[int] = None
    bfd_state: BFDState
    bfd_ix: int
    Neigh_rDNS: str = ""  # ?

    @computed_field  # type: ignore
    @property
    def port(self) -> str:
        return str(self.bfd_addr) if self.bfd_addr else f"ix {self.bfd_ix}"

    @computed_field  # type: ignore
    @property
    def description(self) -> str:
        return f"{self.Neigh_rDNS}, {self.lastevent}"

    @computed_field  # type: ignore
    @property
    def op_state(self) -> str:
        return f"BFD  {self.bfd_state[:5]}"

    def is_down(self) -> bool:
        return self.bfd_state == BFDState.DOWN


class BGPEvent(Event):
    type: str = Event.Type.BGP
    bgp_AS: str
    bgp_OS: str
    remote_AS: int
    remote_addr: IPvAnyAddress
    peer_uptime: int
    lastevent: str

    @computed_field  # type: ignore
    @property
    def port(self) -> str:
        return f"AS{self.remote_AS}"

    @computed_field  # type: ignore
    @property
    def description(self) -> str:
        # rdns = dns_reverse_resolver(str(cls.remote_addr))
        # rdns = ''
        # return f"{rdns}, {cls.lastevent}"
        return f"{self.remote_addr}, {self.lastevent}"

    @computed_field  # type: ignore
    @property
    def op_state(self) -> str:
        return f"BGP  {self.bgp_OS[:5]}"

    def is_down(self) -> bool:
        return self.bgp_OS == "down"


class ReachabilityEvent(Event):
    type: str = Event.Type.REACHABILITY
    reachability: ReachabilityState
    ac_down: Optional[timedelta] = None  # int
    description: str = ""
    port: str = ""

    @computed_field  # type: ignore
    @property
    def op_state(self) -> str:
        return self.reachability

    def is_down(self) -> bool:
        return self.reachability == ReachabilityState.NORESPONSE


class PortStateEvent(Event):
    type: str = Event.Type.PORTSTATE
    ac_down: Optional[timedelta] = None  # int
    descr: str = ""
    flaps: Optional[int] = None
    flapstate: Optional[FlapState] = None
    if_index: int
    port_state: PortState  # str *
    reason: Optional[str] = None  # *
    port: str = ""

    EXTRAS: ClassVar[Set[str]] = {"clear_flapping", "get_downtime"}

    @computed_field  # type: ignore
    @property
    def description(self) -> str:
        return self.descr

    @computed_field  # type: ignore
    @property
    def op_state(self) -> str:
        return f"PORT  {self.port_state[:5]}"

    def get_downtime(self):
        """Calculate downtime on this PortState"""
        # If no transition is detected, use now.
        now = utcnow()
        lasttrans = self.lasttrans or now
        accumulated = self.ac_down or timedelta(seconds=0)

        if self.port_state in [PortState.DOWN, PortState.LOWER_LAYER_DOWN]:
            return accumulated + now - lasttrans
        else:
            return accumulated

    def is_down(self) -> bool:
        return self.port_state in [PortState.DOWN, PortState.LOWER_LAYER_DOWN]



from typing import Optional

from pydantic import BaseModel


class UserConfig(BaseModel):
    username: str
    password: str


class OptionalUserConfig(BaseModel):
    username: Optional[str]
    password: Optional[str]


class ServerV1Config(BaseModel):
    server: str
    port: int = 8001


class Options(BaseModel):
    autoremove: bool = False
    timeout: int = 30



import os
from pathlib import Path
from tempfile import mkstemp, mkdtemp
from unittest import TestCase

from zinolib.config.utils import find_config_file, make_filename_safe

from ..utils import make_tmptextfile, delete_tmpfile


class MakeFilenameSafeTest(TestCase):

    def test_if_no_evil_return_input(self):
        filename = 'foo.bar.xux'
        result = make_filename_safe(filename)
        self.assertEqual(filename, result)

    def test_chop_away_path_bits(self):
        filename = '/etc/otherprogram/foo.bar.xux'
        result = make_filename_safe(filename)
        self.assertEqual(result, 'foo.bar.xux')


class FindConfigFileTest(TestCase):
    def test_find_config_file_missing_config_file(self):
        with self.assertRaises(FileNotFoundError):
            find_config_file("bcekjyfbu eyxxgyikyvub iysbiucbcsiu")

    def test_find_config_file_golden_path(self):
        filename = make_tmptextfile("", None)
        tmp_directory = Path(filename).parent
        found_filename = find_config_file(filename, directories=[tmp_directory])
        delete_tmpfile(filename)
        self.assertEqual(Path.cwd() / filename, found_filename)



import logging
import datetime
import os
import time
import unittest
from ipaddress import ip_address

from zinolib.ritz import ritz, ProtocolError, AuthenticationError, caseState, caseType
from zinolib.zino_emu import zinoemu

from ..utils import executor


class SlowTest(unittest.TestCase):
    def setUp(self):
        self.old_timezone = os.environ.get('TZ', None)
        os.environ['TZ'] = 'Europe/Oslo'
        time.tzset()

    def tearDown(self):
        if self.old_timezone:
            os.environ['TZ'] = self.old_timezone
            time.tzset()

    def test_C_connect_no_login_response(self):
        with zinoemu(executor):

            r = ritz(
                "127.0.0.1", username="no_login_response", password="no_login_response"
            )
            #  with self.assertRaises(ProtocolError):
            try:
                r.connect()
            except ProtocolError:
                pass
            finally:
                r.close()

    def test_E_connect_random_data_on_connect(self):
        with zinoemu(executor):
            r = ritz(
                "127.0.0.1",
                username="illegal-first-response",
                password="illegal-first-response",
            )
            with self.assertRaises(ProtocolError):
                r.connect()
            r.close()



from pathlib import Path
import re

from .utils import find_config_file


TCL_FILENAME = ".ritz.tcl"


def load(filename):
    """Get the contents of a .ritz.tcl config file

    .ritz.tcl is formatted as a tcl file.
    """
    path = Path(filename).expanduser()
    with path.open("r") as f:
        # normalize lineends just in case
        lines = [line.strip() for line in f.readlines()]
    text = "\n".join(lines)
    return text


def parse(text):
    """
    Parse the text of a .ritz.tcl config file

    .ritz.tcl is formatted as a tcl file.

    A config-file with the following contents:

    set Secret 0123456789
    set User admin
    set Server example.org
    set Port 8001

    global Sortby
    set Sortby "upd-rev"

    set _Secret(dev-server) 0123456789
    set _User(dev-server) admin
    set _Server(dev-server) example.com
    set _Port(dev-server) 8001

    Results in the following dict:

    {
        "default": {
            "Secret": "0123456789",
            "User": "admin",
            "Server": "example.org",
            "Port": "8001",
            "Sortby": "upd-rev",
        },
        "dev-server": {
            "Secret": "0123456789",
            "User": "admin",
            "Server": "example.com",
            "Port": "8001",
        },
    }
    """
    lines = text.split("\n")
    config = {}
    for line in lines:
        match = re.fullmatch(r"\s?set _?([a-zA-Z0-9]+)(?:\((.*)\))? (.*)", line)
        if match:
            key, group, value = match.groups()
            if not group:
                group = "default"

            if group not in config:
                config[group] = {}

            config[group][key] = value
    return config


def normalize(tcl_config_dict):
    """
    Standardize on snake-case key-names and break out global options

    Usage::

        > config_dict = Normalizer.normalize(tcl_config_dict)

    A config-dict with the following contents::

        {
            "default": {
                "Secret": "0123456789",
                "User": "admin",
                "Server": "example.org",
                "Port": "8001",
                "Sortby": "upd-rev",
            },
            "dev-server": {
                "Secret": "0123456789",
                "User": "admin",
                "Server": "example.com",
                "Port": "8001",
            },
        }

    Results in a dict with the following contents::

        {
            "connections": {
                "default": {
                    "secret": "0123456789",
                    "username": "admin",
                    "server": "example.org",
                    "port": "8001",
                },
                "dev_server": {
                    "secret": "0123456789",
                    "username": "admin",
                    "server": "example.com",
                    "port": "8001",
                },
            },
            "global_options": {
                "sort_by": "upd-rev",
            },
        }
    """
    KEYMAP = {"Sortby": "sort_by", "User": "username"}
    CONNECTION_KEYS = set(("username", "secret", "server", "port"))
    connections = {}
    global_options = {}
    for name in tcl_config_dict:
        connection = {}
        for key, value in tcl_config_dict[name].items():
            key = KEYMAP.get(key, key.lower())
            key = "_".join(key.split("-"))
            if key not in CONNECTION_KEYS:
                global_options[key] = value
            else:
                connection[key] = value
        name = "_".join(name.split("-"))
        connections[name] = connection
    return {"global_options": global_options, "connections": connections}


# legacy
def parse_tcl_config(filename=None):
    if filename is None:
        filename = TCL_FILENAME
    filename = find_config_file(filename)
    contents = load(filename)
    config = parse(contents)
    return config



import unittest
from unittest import mock
from unittest.mock import create_autospec
from datetime import datetime, timedelta

from pydantic import ValidationError

from zinolib.controllers.base import EventManager
from zinolib.event_types import Event, HistoryEntry, LogEntry
from zinolib.event_types import AdmState, BFDState, PortState, ReachabilityState


common_minimal_input = {
    "id": 4576,
    "adm_state": AdmState.OPEN,
    "router": "fghjuy",
    "opened": 1000000000,
}


class EventTest(unittest.TestCase):

    def test_create_event_from_empty_dict_should_complain_about_missing_type(self):
        with self.assertRaises(KeyError) as cm:
            event = Event.create({})
        self.assertEqual(cm.exception.args[0], "type")

    def test_create_event_from_minimal_dict_should_fail(self):
        minimal_input = common_minimal_input.copy()
        minimal_input["type"] = Event.Type.ALARM
        with self.assertRaises(ValidationError):
            event = Event.create(minimal_input)

    def test_create_event_from_subtype_correct_dict_should_succeed(self):
        minimal_input = common_minimal_input.copy()
        minimal_input.update(**{
            "type": Event.Type.ALARM,
            "alarm_count": 1,
            "alarm_type": "apocalypse!!",
        })
        event = Event.create(minimal_input)
        self.assertTrue(isinstance(event, Event.SUBTYPES[Event.Type.ALARM.value]))

    def test_get_downtime_should_fail_noisily_on_non_PortStateEvent(self):
        minimal_input = common_minimal_input.copy()
        minimal_input.update(**{
            "type": Event.Type.ALARM,
            "alarm_count": 1,
            "alarm_type": "apocalypse!!",
        })
        event = Event.create(minimal_input)
        with self.assertRaises(AttributeError):
            event.get_downtime()

    def test_ac_down_field_should_be_timedelta(self):
        # pydantic changes its mind as to what formats are supported
        # so we double check here
        minimal_input = common_minimal_input.copy()
        minimal_input.update(**{
            "type": Event.Type.REACHABILITY,
            "reachability": ReachabilityState.REACHABLE,
            "ac_down": 4564654,
        })
        event = Event.create(minimal_input)
        self.assertTrue(isinstance(event.ac_down, timedelta))

    def test_lasttrans_and_updated_should_be_datetime(self):
        # pydantic changes its mind as to what formats are supported
        # so we double check here
        input = common_minimal_input.copy()
        input.update(**{
            "type": Event.Type.REACHABILITY,
            "reachability": ReachabilityState.REACHABLE,
            "ac_down": 4564654,
            "lasttrans": "1234567890",
            "updated": "1234567890",
        })
        event = Event.create(input)
        self.assertTrue(isinstance(event.lasttrans, datetime))
        self.assertTrue(isinstance(event.updated, datetime))


class AlarmEventTest(unittest.TestCase):

    def test_create_alarm_event_sets_correct_fields(self):
        minimal_input = common_minimal_input.copy()
        minimal_input.update(**{
            "type": Event.Type.ALARM,
            "alarm_count": 1,
            "alarm_type": "apocalypse!!",
            "lastevent": "vfbghj",
        })
        event = Event.create(minimal_input)
        self.assertEqual(event.alarm_count, minimal_input["alarm_count"])
        self.assertEqual(event.alarm_type, minimal_input["alarm_type"])
        self.assertEqual(event.op_state, f"ALRM  {event.alarm_type}")
        self.assertEqual(event.description, event.lastevent)


class BFDEventTest(unittest.TestCase):

    def test_create_bfd_event_sets_correct_fields(self):
        minimal_input = common_minimal_input.copy()
        minimal_input.update(**{
            "type": Event.Type.BFD,
            "bfd_state": BFDState.UP,
            "bfd_ix": 4356,
            "lastevent": "vfbghj",
        })
        event = Event.create(minimal_input)
        self.assertEqual(event.port, f"ix {minimal_input['bfd_ix']}")
        self.assertEqual(event.op_state, f"BFD  {event.bfd_state}"[:10])
        self.assertEqual(event.description, f", {minimal_input['lastevent']}")


class BGPEventTest(unittest.TestCase):

    def test_create_bgp_event_sets_correct_fields(self):
        minimal_input = common_minimal_input.copy()
        minimal_input.update(**{
            "type": Event.Type.BGP,
            "bgp_AS": "vfrgthyj",
            "bgp_OS": "hjkuil",
            "remote_AS": 345,
            "remote_addr": "8.8.8.8",
            "peer_uptime": 4567,
            "lastevent": "ghj",
        })
        event = Event.create(minimal_input)
        self.assertEqual(event.port, f"AS{minimal_input['remote_AS']}")
        self.assertEqual(event.op_state, f"BGP  {event.bgp_OS}"[:10])
        self.assertEqual(event.description, f"{minimal_input['remote_addr']}, {minimal_input['lastevent']}")


class ReachabilityEventTest(unittest.TestCase):

    def test_create_reachability_event_sets_correct_fields(self):
        minimal_input = common_minimal_input.copy()
        minimal_input.update(**{
            "type": Event.Type.REACHABILITY,
            "reachability": ReachabilityState.REACHABLE,
            "ac_down": 567,
        })
        event = Event.create(minimal_input)
        self.assertEqual(event.port, "")
        self.assertEqual(event.op_state, event.reachability)
        self.assertEqual(event.description, "")


class PortStateEventTest(unittest.TestCase):

    def test_create_port_state_event_sets_correct_fields(self):
        minimal_input = common_minimal_input.copy()
        minimal_input.update(**{
            "type": Event.Type.PORTSTATE,
            "if_index": 321,
            "port_state": PortState.UP,
        })
        event = Event.create(minimal_input)
        self.assertEqual(event.port, "")
        self.assertEqual(event.op_state, f"PORT  {event.port_state}"[:11])
        self.assertEqual(event.description, event.descr)

    def test_get_downtime_should_return_a_timedelta(self):
        minimal_input = common_minimal_input.copy()
        minimal_input.update(**{
            "type": Event.Type.PORTSTATE,
            "if_index": 321,
            "port_state": PortState.UP,
        })
        event = Event.create(minimal_input)
        downtime = event.get_downtime()
        self.assertTrue(isinstance(downtime, timedelta))


class HistoryEntryTest(unittest.TestCase):

    def test_create_list_should_return_list_of_history_entries(self):
        dt = datetime.fromisoformat("2023-06-28T10:41:54+00:00")
        history_list = [{
            "date": dt,
            "log": "fhgj",
            "user": "ghj",
        }]
        history = HistoryEntry.create_list(history_list)
        expected_history_entry = HistoryEntry(date=dt, log="fhgj", user="ghj")
        self.assertEqual(history, [expected_history_entry])


class LogEntryTest(unittest.TestCase):

    def test_create_list_should_return_list_of_log_entries(self):
        dt = datetime.fromisoformat("2023-06-28T10:41:54+00:00")
        log_list = [{
            "date": dt,
            "log": "fhgj",
        }]
        log = LogEntry.create_list(log_list)
        expected_log_entry = LogEntry(date=dt, log="fhgj")
        self.assertEqual(log, [expected_log_entry])


class EventManagerTest(unittest.TestCase):

    def setUp(self):
        minimal_input = common_minimal_input.copy()
        minimal_input.update(**{
            "type": Event.Type.ALARM,
            "alarm_count": 1,
            "alarm_type": "apocalypse!!",
            "lastevent": "vfbghj",
        })
        event = Event.create(minimal_input)
        self.event = event

    def test__verify_session_wjen_no_session_should_fail_noisily(self):
        event_manager = EventManager()
        with self.assertRaises(ValueError):
            event_manager._verify_session()

    def test_get_event_with_id_should_succeed(self):
        event = self.event
        event_manager = EventManager()
        event_manager._set_event(event)

        resulting_event = event_manager._get_event(event.id)
        self.assertEqual(self.event, resulting_event)

    def test_set_history_for_event(self):
        event = self.event
        event_manager = EventManager()
        event_manager._set_event(event)
        self.assertFalse(event_manager.events[event.id].history)
        dt = datetime.fromisoformat("2023-06-28T10:41:54+00:00")
        history_list = [{
            "date": dt,
            "log": "fhgj",
            "user": "ghj",
        }]
        updated_event = event_manager.set_history_for_event(event, history_list)
        self.assertTrue(event_manager.events[event.id].history)

    def test_set_log_for_event(self):
        event = self.event
        event_manager = EventManager()
        event_manager._set_event(event)
        self.assertFalse(event_manager.events[event.id].log)
        dt = datetime.fromisoformat("2023-06-28T10:41:54+00:00")
        log_list = [{
            "date": dt,
            "log": "fhgj",
        }]
        updated_event = event_manager.set_log_for_event(event, log_list)
        self.assertTrue(event_manager.events[event.id].log)


class AdmStateTest(unittest.TestCase):

    def test_golden_path(self):
        for state in AdmState:
            enum_state = AdmState(state)
            self.assertEqual(enum_state.value, state)

    def test_garbage_input_should_be_converted_to_UNKNOWN(self):
        value = AdmState("random garbage")
        self.assertEqual(value, AdmState.UNKNOWN)
